<?php

namespace Drupal\monarch_entity_yml;

use Drupal\Core\Config\FileStorage;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityRepositoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\path\Plugin\Field\FieldType\PathFieldItemList;

/**
 * Defines import actions for the install profile.
 */
class Import {

  /**
   * The entity type manager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The entity repository service.
   *
   * @var \Drupal\Core\Entity\EntityRepositoryInterface
   */
  protected $entityRepository;

  /**
   * The entity field manager service.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * Error handling mode.
   *
   * @var string
   */
  public $mode = 'throw';

  /**
   * Create an Import helper class.
   */
  public function __construct(EntityTypeManagerInterface $entityTypeManager, EntityRepositoryInterface $entityRepository, EntityFieldManagerInterface $entityFieldManager) {
    $this->entityTypeManager = $entityTypeManager;
    $this->entityRepository = $entityRepository;
    $this->entityFieldManager = $entityFieldManager;
  }

  /**
   * Import pre-parsed entity data as entities.
   *
   * @param string $target_uuid
   *   The UUID to import.
   * @param array $data
   *   The pre-parsed entity data.
   * @param array $feedback
   *   Existing array of entity descriptors.
   * @param array $done
   *   List of entity uuids already imported.
   *
   * @return array
   *   Array of entity descriptors created.
   */
  public function importData(string $target_uuid = NULL, array $data = [], array &$feedback = [], array &$done = []) {
    foreach ($data as $definition) {
      if (!isset($definition['storage']) || !is_string($definition['storage'])) {
        throw new \Exception('Storage not specified for entity entry.');
      }

      if (!isset($definition['values']) || !is_array($definition['values']) || empty($definition['values'])) {
        throw new \Exception('No valid values for entity entry.');
      }

      $entity_type_id = $definition['storage'];
      $uuid = $this->getuuid($definition);

      if (!$uuid || !is_string($uuid)) {
        throw new \Exception('UUID not specified for entity entry.');
      }

      if (!empty($done[$uuid])) {
        continue;
      }

      if ($target_uuid && $target_uuid !== $uuid) {
        continue;
      }

      $done[$uuid] = $uuid;

      try {
        foreach ($definition['dependencies'] ?? [] as $dependency_uuid) {
          $this->importData($dependency_uuid, $data, $feedback, $done);
        }

        $entity_definition = $this->entityTypeManager->getDefinition($entity_type_id);
        $lookup = [];
        $entity = NULL;
        $revisionable = $entity_definition->isRevisionable();

        foreach ($entity_definition->getKeys() as $key_type => $key) {
          if ($key && in_array($key_type, [
            'id',
            'revision',
            'uuid',
          ]) && ($definition['values'][$key] ?? NULL)) {
            $lookup[$key_type] = $definition['values'][$key];

            while (is_array($lookup[$key_type])) {
              $lookup[$key_type] = reset($lookup[$key_type]);
            }
          }
        }

        $bundle_key = $entity_definition->getKeys()['bundle'] ?? NULL;
        $bundle = NULL;

        if ($bundle_key) {
          $bundle = $definition['values'][$bundle_key] ?? NULL;

          while (is_array($bundle)) {
            $bundle = reset($bundle);
          }
        }

        if (empty($lookup)) {
          throw new \Exception('No valid lookup!');
        }

        if (isset($lookup['revision'])) {
          /** @var \Drupal\Core\Entity\ContentEntityInterface $entity */
          $entity = $this->entityTypeManager->getStorage($entity_type_id)->loadRevision($lookup['revision']);
        }
        elseif (isset($lookup['id'])) {
          /** @var \Drupal\Core\Entity\ContentEntityInterface $entity */
          $entity = $this->entityTypeManager->getStorage($entity_type_id)->load($lookup['id']);
        }
        elseif (isset($lookup['uuid'])) {
          /** @var \Drupal\Core\Entity\ContentEntityInterface $entity */
          $entity = $this->entityRepository->loadEntityByUuid($entity_type_id, $lookup['uuid']);
        }

        $field_definitions = $this->entityFieldManager->getFieldDefinitions($entity_type_id, $bundle);

        foreach ($field_definitions as $field_name => $field_definition) {
          if (in_array($field_definition->getType(), [
            'entity_reference',
            'entity_reference_revisions',
          ])) {
            $target_entity_type_id = $field_definition->getSettings()['target_type'] ?? NULL;

            foreach ($definition['values'][$field_name] ?? [] as &$field_value) {
              if ($target_uuid = ($field_value['target_uuid'] ?? NULL)) {
                $target_entity_definition = $this->entityTypeManager->getDefinition($target_entity_type_id);
                if ($uuidKey = ($target_entity_definition->getKeys()['uuid'] ?? NULL)) {
                  $id = current($this->entityTypeManager->getStorage($target_entity_type_id)->getQuery()->condition($uuidKey, $target_uuid)->execute());

                  if (!is_bool($id)) {
                    $field_value['target_id'] = $id;
                    unset($field_value['target_uuid']);
                  }
                }
              }
            }
          }
        }

        if (!$entity) {
          $created = TRUE;

          /** @var \Drupal\Core\Entity\ContentEntityInterface $entity */
          $entity = $this->entityTypeManager->getStorage($entity_type_id)->create($definition['values']);
        }
        else {
          $created = FALSE;
        }

        /** @var \Drupal\Core\Entity\ContentEntityInterface $entity */
        if ($entity) {
          $path_fields = [];
          /** @var \Drupal\Core\Entity\ContentEntityInterface $entity */
          foreach ($definition['values'] as $key => $value) {
            if ($entity->get($key) instanceof PathFieldItemList) {
              $path_fields[$key] = $value;
            }
            else {
              $entity->get($key)->setValue($value);
            }
          }

          if ($revisionable) {
            $entity->setNewRevision(FALSE);
          }
  
          $entity->setSyncing(TRUE);
          $entity->save();

          if (!empty($path_fields)) {
            // Reload so we can detect any new pids if they exist.
            if (isset($lookup['revision'])) {
              /** @var \Drupal\Core\Entity\ContentEntityInterface $entity */
              $entity = $this->entityTypeManager->getStorage($entity_type_id)->loadRevision($lookup['revision']);
            }
            elseif (empty($entity) && isset($lookup['id'])) {
              /** @var \Drupal\Core\Entity\ContentEntityInterface $entity */
              $entity = $this->entityTypeManager->getStorage($entity_type_id)->load($lookup['id']);
            }
            elseif (empty($entity) && isset($lookup['uuid'])) {
              /** @var \Drupal\Core\Entity\ContentEntityInterface $entity */
              $entity = $this->entityRepository->loadEntityByUuid($entity_type_id, $lookup['uuid']);
            }
    
            /** @var \Drupal\Core\Entity\ContentEntityInterface $entity */
            foreach ($path_fields as $path_field => $path_field_data) {
              $pathauto = $path_field_data['pathauto'] ?? NULL;
              $old_path_data = $entity->{$path_field}->getValue();
              unset($path_field_data['pathauto']);

              foreach ($old_path_data as $index => $values) {
                if (isset($values['pid']) && isset($path_field_data[$index]['alias'])) {
                  $path_field_data[$index]['pid'] = $values['pid'];
                }
              }

              $entity->{$path_field}->setValue($path_field_data);

              if (!is_null($pathauto)) {
                $entity->{$path_field}->pathauto = $pathauto;
              }
            }

            if ($revisionable) {
              $entity->setNewRevision(FALSE);
            }

            $entity->setSyncing(TRUE);
            $entity->save();
          }
        }

        $feedback[] = '[' . ($entity->uuid() ?? 'No UUID') . ', ' . ($created ? 'create' : 'update') . '] ' . $entity->getEntityTypeId();
      }
      catch (\Throwable $e) {
        if ($entity) {
          $msg = 'Error importing ' . $entity_type_id . ':' . ($definition['values']['uuid'][0]['value'] ?? 'no_UUID') . ' - ' . $e->getMessage();

          if ($this->mode === 'continue') {
            $feedback[] = $msg;
            $feedback[] = 'Continuing despite error...';
          }
          else {
            throw new \Exception($msg);
          }
        }
        else {
          throw $e;
        }
      }
    }

    return $feedback;
  }

  /**
   * Get uuid from an entry.
   */
  protected function getuuid(array $data) {
    return $data['values']['uuid'][0]['value'] ?? $data['values']['uuid']['value'] ?? $data['values']['uuid'] ?? $data['uuid'] ?? NULL;
  }

  /**
   * Generates baseline content (intended for after config-import).
   *
   * @param string $path
   *   The path to import files from.
   * @param array $feedback
   *   Existing array of entity descriptors.
   *
   * @return array
   *   Array of entity descriptors created.
   */
  public function import(string $path, array &$feedback = []) {
    $storage = new FileStorage(
      $path
    );

    $files = $storage->readMultiple($storage->listAll(''));
    $feedback = [];

    $full_data = [];

    foreach ($files as $filename => $data) {
      if ($filename[0] !== '_') {
        if (isset($data['storage'])) {
          $uuid = $this->getuuid($data);

          if ($uuid && is_string($uuid)) {
            $full_data[$uuid] = $data;
          }
        }
        else {
          foreach ($data as $entry) {
            if (is_string($entry)) {
              $entry = $files[$entry] ?? [];
            }

            $uuid = $this->getuuid($entry);

            if ($uuid && is_string($uuid)) {
              $full_data[$uuid] = $entry;
            }
          }
        }
      }
    }

    foreach (array_keys($full_data) as $uuid) {
      if ($full_data[$uuid]['storage'] === 'path_alias') {
        $this->importData($uuid, $full_data, $feedback);
      }
    }

    foreach (array_keys($full_data) as $uuid) {
      if ($full_data[$uuid]['storage'] !== 'path_alias') {
        $this->importData($uuid, $full_data, $feedback);
      }
    }

    return $feedback;
  }

}
